/*
WBLoader 1.0

Demonstrates the use of WBLoader firmware as a simple loader for communication through a phoenix interface.

Both Infinity USB Unlimited SDK and a standard phoenix interfaces can be used.
*/

//Define 'USE_STDPHOENIX', if you wish to compile with use for a standard phoenix interface
//or Define 'USE_INFUNLTDSDK' to compile with use for the Infinity USB Unlimited SDK

//#define USE_STDPHOENIX
#define USE_INFUNLTDSDK

//Define a COM port to use if you have specified USE_STDPHOENIX
#define COM_PORT "\\\\.\\COM1"

#include <windows.h>
#include <stdio.h>
#include <conio.h>

#ifdef USE_INFUNLTDSDK
#include "infunltd.h"
#endif

//Defines the page sizes of the loader
#define ReadPageSize 16
#define WritePageSize 16

//The total amount of pages depends on the size of the EEprom used
#define TotalPages 512	//16x512 = 8192 > 24c64
//#define TotalPages 8192	//16x8192 = 131072 > 24c1024


//Function prototypes
HANDLE OpenDevice();
int CloseDevice(HANDLE hDevice);
int InitPhoenix(HANDLE hDevice);
int WritePage(HANDLE hDevice, unsigned char* outdata, unsigned int iPage);
int ReadPage(HANDLE hDevice, unsigned char* indata, unsigned int iPage);
int Phoenix_WaitForBytesTimeOut(HANDLE hDevice, unsigned int iNumberOfBytes, DWORD dwMillisecs);
unsigned short int CalcCRC16(unsigned char bytedata, unsigned short int oldCRCvalue);

/*
Waits for the specified amount of bytes to be ready, returns when the bytes are ready (0) or a timeout occurs (-1)
*/
int Phoenix_WaitForBytesTimeOut(HANDLE hDevice, unsigned int iNumberOfBytes, DWORD dwMillisecs)
{
	long Timer1 = GetTickCount();

#ifdef USE_INFUNLTDSDK

	//Loop until specified amount of data is ready, or the time has expired
	while(INFUNLTD_Phoenix_BytesInFifo(hDevice) < iNumberOfBytes)
	{
		Sleep(10);
		if(GetTickCount() > Timer1 + dwMillisecs) return -1;
	}

#else

#ifdef USE_STDPHOENIX
	COMSTAT comStat;
	DWORD   dwErrors;
	
	ClearCommError(hDevice, &dwErrors, &comStat);

	//Loop until specified amount of data is ready, or the time has expired
	while(comStat.cbInQue < iNumberOfBytes)
	{
		Sleep(10);
		if(GetTickCount() > Timer1 + dwMillisecs) return -1;
		ClearCommError(hDevice, &dwErrors, &comStat);
	}
	
#endif

#endif

	return 0;
}

/*
Routine used to calculate CRC on data
*/
unsigned short int CalcCRC16(unsigned char bytedata, unsigned short int oldCRCvalue)
{
	unsigned short int crc = oldCRCvalue;

	crc  = (unsigned char)(crc >> 8) | (crc << 8);
	crc ^= bytedata;
	crc ^= (unsigned char)(crc & 0xff) >> 4;
	crc ^= (crc << 8) << 4;
	crc ^= ((crc & 0xff) << 4) << 1;

	return crc;
}

/*
Opens the device (Either Infinity USB Unlimited if USE_INFUNLTDSDK is defined, or a COM port if USE_STDPHOENIX is defined at compile time
*/
HANDLE OpenDevice()
{
	HANDLE hDevice = 0;

#ifdef USE_INFUNLTDSDK
	DWORD dwDevice = 0;
	SDK_STATUS sdk_status = SDK_SUCCESS;

	//Open the selected device number, the handle hDevice will receive the handle
	sdk_status = INFUNLTD_OpenDeviceFromNum(dwDevice, &hDevice, 0);

	//Set USB communication timeouts
	sdk_status = INFUNLTD_SetTimeouts(2000,2000);
#else

#ifdef USE_STDPHOENIX
	//Create a handle to the standard COM port
	hDevice = CreateFile(COM_PORT,GENERIC_READ|GENERIC_WRITE,FILE_SHARE_READ|FILE_SHARE_WRITE,0,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
#endif

#endif

	return hDevice;
}

/*
Closes the device and shuts down phoenix mode
*/
int CloseDevice(HANDLE hDevice)
{
#ifdef USE_INFUNLTDSDK
	SDK_STATUS sdk_status = SDK_SUCCESS;

	//Disable phoenix mode
	sdk_status = INFUNLTD_Phoenix_Disable(hDevice);

	//Close down the frequency generator by setting it to 0
	sdk_status = INFUNLTD_Phoenix_SetFrequency(hDevice,0);

	//Close the handle
	sdk_status = INFUNLTD_CloseDevice(hDevice);
#else

#ifdef USE_STDPHOENIX
	//Close the handle
	CloseHandle(hDevice);
#endif

#endif

	return 0;
}

/*
Initializes phoenix mode, resets the card and reads an ATR
*/
int InitPhoenix(HANDLE hDevice)
{
	unsigned char ATR[255];
	DWORD dwRead = 0;
	
	memset(ATR,0,sizeof(ATR));

#ifdef USE_INFUNLTDSDK
	SDK_STATUS sdk_status = SDK_SUCCESS;
	
	//Open phoenix mode 9600E28
	sdk_status = INFUNLTD_Phoenix_Enable(hDevice,1,2);

	//Set frequency to 3.579MHz
	sdk_status = INFUNLTD_Phoenix_SetFrequency(hDevice,3579000);

	//Set card voltage to 5v
	sdk_status = INFUNLTD_SetVoltage(hDevice,5);
	
	//Reset card
	sdk_status = INFUNLTD_Phoenix_ResetCard(hDevice);

	//Wait for ATR
	Sleep(200);

	//Get the ATR length
	int ATRLen = INFUNLTD_Phoenix_BytesInFifo(hDevice);

	//Read ATR
	INFUNLTD_Phoenix_Read(hDevice, ATR, ATRLen, &dwRead);

	//Display ATR
	printf("Card ATR = %s\n\n",ATR);

#else

#ifdef USE_STDPHOENIX
	DCB dcb = {0};
	dcb.DCBlength = sizeof(dcb);

	if (!GetCommState(hDevice, &dcb))
	{
		CloseHandle(hDevice);
		return -1;
	}

	dcb.BaudRate = 9600;
	dcb.ByteSize = 8;
	dcb.Parity = EVENPARITY;
	dcb.StopBits = TWOSTOPBITS;
	dcb.EvtChar = '\0';

	dcb.fDtrControl = 0;
	dcb.fRtsControl = 0;

	dcb.fOutxCtsFlow = 0;
	dcb.fOutxDsrFlow = 0;
	dcb.fDsrSensitivity = 0;
	dcb.fOutX = 0;
	dcb.fInX = 0;
	dcb.fTXContinueOnXoff = 0;
	dcb.XonChar = 0;
	dcb.XoffChar = 0;
	dcb.XonLim = 0;
	dcb.XoffLim = 0;

	//dcb.ErrorChar = 'X';
	dcb.fParity = TRUE;

	if (!SetCommState(hDevice, &dcb))
	{
		CloseHandle(hDevice);
		return -2;
	}

	//Reset card
	EscapeCommFunction(hDevice,SETRTS);
	Sleep(50);
	EscapeCommFunction(hDevice,CLRRTS);

	//Wait for ATR
	Sleep(200);

	//Get the ATR length
	COMSTAT comStat;
	DWORD   dwErrors;
	ClearCommError(hDevice, &dwErrors, &comStat);
	int ATRLen = comStat.cbInQue;

	//Read ATR
	ReadFile(hDevice, ATR, ATRLen, &dwRead, 0);

	//Display ATR
	printf("Card ATR = %s\n\n",ATR);

#endif

#endif

	//Return 0 if no error
	return 0;
}

/*
Writes a page of data to the card
*/
int WritePage(HANDLE hDevice, unsigned char* outdata, unsigned int iPage)
{
	DWORD dwWritten = 0;
	DWORD dwRead = 0;

	unsigned short int CRCval = 0xFFFF;
	unsigned char buffer[WritePageSize + 6];
	unsigned short int WriteRes = 0;

	//Protocol for writing a page = <CMD,High Adr,Low Adr,Data>
	buffer[0] = 0x03;
	buffer[1] = HIBYTE(iPage);
	buffer[2] = LOBYTE(iPage);

	//Copy the outgoing data to the write buffer
	memcpy(buffer+3,outdata,WritePageSize);

	//Calculate CRC for the data
	CRCval = 0xFFFF;
	for(int i = 3; i < 3+ReadPageSize; i++)
	{
		CRCval = CalcCRC16(buffer[i], CRCval);
	}

	//Append the CRC value to the outgoing buffer
	buffer[WritePageSize + 3] = HIBYTE(CRCval);
	buffer[WritePageSize + 4] = LOBYTE(CRCval);

#ifdef USE_INFUNLTDSDK

	//Write the outgoing buffer to the device
	INFUNLTD_Phoenix_Write(hDevice, buffer,WritePageSize + 5, &dwWritten);
	if(dwWritten != ReadPageSize + 5) return -3;

	//Wait for the echo + response
	int status = Phoenix_WaitForBytesTimeOut(hDevice, ReadPageSize + 5 + 1, 2000);
	if(status != 0) return -5;

	//Read the echo + response
	INFUNLTD_Phoenix_Read(hDevice, buffer, ReadPageSize + 5 + 1, &dwRead);
	if(dwRead != ReadPageSize + 5 + 1) return -4;

	//Get the response
	WriteRes = buffer[21];

	//Check the response for errors
	if(WriteRes == 1) return -1;	//CRC error
	if(WriteRes == 2) return -2;	//Error writing EEprom

	#else
	#ifdef USE_STDPHOENIX

	//Write the outgoing buffer to the device
	WriteFile(hDevice, buffer,WritePageSize + 5, &dwWritten, 0);
	if(dwWritten != ReadPageSize + 5) return -3;

	//Wait for the echo + response
	int status = Phoenix_WaitForBytesTimeOut(hDevice, ReadPageSize + 5 + 1, 2000);
	if(status != 0) return -5;

	//Read the echo + response
	ReadFile(hDevice, buffer, ReadPageSize + 5 + 1, &dwRead, 0);
	if(dwRead != ReadPageSize + 5 + 1) return -4;

	//Get the response
	WriteRes = buffer[21];

	//Check the response for errors
	if(WriteRes == 1) return -1;	//CRC error
	if(WriteRes == 2) return -2;	//Error writing EEprom

	#endif
	#endif

	//Return 0 if no error
	return 0;
}

/*
Reads a page of data from the card
*/
int ReadPage(HANDLE hDevice, unsigned char* indata, unsigned int iPage)
{
	DWORD dwRead = 0;
	DWORD dwWritten = 0;

	unsigned short int CRCval = 0xFFFF;
	unsigned short int CRCret = 0;
	unsigned char buffer[ReadPageSize + 2];

	//Protocol for reading a page = <CMD,High Adr,Low Adr>
	buffer[0] = 0x02;
	buffer[1] = HIBYTE(iPage);
	buffer[2] = LOBYTE(iPage);

	#ifdef USE_INFUNLTDSDK

	//Write the command to the device
	INFUNLTD_Phoenix_Write(hDevice, buffer, 3, &dwWritten);
	if(dwWritten != 3) return -1;

	//Read the echo
	INFUNLTD_Phoenix_Read(hDevice, buffer, 3, &dwRead);
	if(dwRead != 3) return -2;

	//Wait for the requested data to arrive
	int status = Phoenix_WaitForBytesTimeOut(hDevice, ReadPageSize + 2, 2000);
	if(status != 0) return -5;

	//Read the requested page data
	INFUNLTD_Phoenix_Read(hDevice, buffer, ReadPageSize + 2, &dwRead);
	if(dwRead != ReadPageSize + 2) return -3;
	
	#else
	#ifdef USE_STDPHOENIX

	//Write the command to the device
	WriteFile(hDevice, buffer, 3, &dwWritten, 0);
	if(dwWritten != 3) return -1;

	//Read the echo
	ReadFile(hDevice, buffer, 3, &dwRead, 0);
	if(dwRead != 3) return -2;

	//Wait for the requested data to arrive
	int status = Phoenix_WaitForBytesTimeOut(hDevice, ReadPageSize + 2, 2000);
	if(status != 0) return -5;

	//Read the requested page data
	ReadFile(hDevice, buffer, ReadPageSize + 2, &dwRead, 0);
	if(dwRead != ReadPageSize + 2) return -3;

	#endif
	#endif

	//Read the CRC from the device
	CRCret = buffer[ReadPageSize];
	CRCret <<= 8;
	CRCret |= buffer[ReadPageSize + 1];

	//Copy data to destination buffer
	memcpy(indata,buffer,ReadPageSize);

	//Calculate CRC on data
	for(int i = 0; i < ReadPageSize; i++)
	{
		CRCval = CalcCRC16(buffer[i], CRCval);
	}

	//Compare read CRC with calculated CRC
	if(CRCval != CRCret)
		return -4;

	//Return 0 if no error
	return 0;
};

int main(int argc, CHAR* argv[])
{
	//prepare a buffer for reading and writing a page
	unsigned char ReadBuffer[ReadPageSize];
	unsigned char WriteBuffer[WritePageSize];
	memset(WriteBuffer,0,WritePageSize);
	memset(ReadBuffer,0,ReadPageSize);

	#ifdef USE_INFUNLTDSDK
	printf("WBLoader 1.0 example, using Infinity USB Unlimited SDK\n\n");
	#else
	#ifdef USE_STDPHOENIX
	printf("WBLoader 1.0 example, using phoenix interface\n\n");
	#endif
	#endif

	//Open the device
	HANDLE hDevice = OpenDevice();
	if(hDevice==INVALID_HANDLE_VALUE)
	{
		printf("Error opening device\nPress any key to exit...\n");
		while(!kbhit());
		return 0;
	}

	//Initialize phoenix mode
	int Res1 = InitPhoenix(hDevice);
	if(Res1 != 0)
	{
		printf("Error initializing phoenix mode\nPress any key to exit...\n");
		while(!kbhit());
		return 0;
	}

	//Write and read the specified amount of pages
	for(int l = 0; l < TotalPages; l++)
	{
		//Fill write buffer with data
		memset(WriteBuffer,'a' + (l % 25),WritePageSize - 1);

		//Write the specific page
		int Res2 = WritePage(hDevice,WriteBuffer,l);
		if(Res2 != 0)
		{
			printf("Error writing page\nPress any key to exit...\n");
			while(!kbhit());
			return 0;
		}
		printf("Wrote page %d / %d = %s\n",l,TotalPages - 1,WriteBuffer);
	
		memset(ReadBuffer,0,ReadPageSize);

		//Read back the same page
		int Res3 = ReadPage(hDevice,ReadBuffer,l);
		if(Res3 != 0)
		{
			printf("Error reading page\nPress any key to exit...\n");
			while(!kbhit());
			return 0;
		}
		printf("Read page %d / %d = %s\n\n",l,TotalPages - 1,ReadBuffer);
	}

	//Close down the device
	CloseDevice(hDevice);

	printf("Press any key to exit...");
	while(!kbhit());

	return 0;
}

